
package com.werken.forehead;

/*
 $Id: ForeheadClassLoader.java,v 1.2 2002/07/27 22:15:45 jvanzyl Exp $

 Copyright 2001 (C) The Werken Company. All Rights Reserved.
 
 Redistribution and use of this software and associated documentation
 ("Software"), with or without modification, are permitted provided
 that the following conditions are met:

 1. Redistributions of source code must retain copyright
    statements and notices.  Redistributions must also contain a
    copy of this document.
 
 2. Redistributions in binary form must reproduce the
    above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other
    materials provided with the distribution.
 
 3. The name "Forehead" must not be used to endorse or promote
    products derived from this Software without prior written
    permission of The Werken Company.  For written permission,
    please contact bob@werken.com.
 
 4. Products derived from this Software may not be called "Forehead"
    nor may "Forehead" appear in their names without prior written
    permission of The Werken Company. Forehead is a registered
    trademark of The Werken Company.
 
 5. Due credit should be given to the Forehead Project
    (http://drools.org/).
 
 THIS SOFTWARE IS PROVIDED BY THE WERKEN COMPANY AND CONTRIBUTORS
 ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 THE WERKEN COMPANY OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.
 
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StringWriter;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

/** Named, dynamically modifiable ClassLoader implementation.
 *
 *  @see Forehead
 *
 *  @author <a href="mailto:bob@eng.werken.com">bob mcwhirter</a>
 *  @author <a href="mailto:jason@zenplex.com">Jason van Zyl</a>
 */
public class ForeheadClassLoader extends URLClassLoader
{
    // ------------------------------------------------------------
    //     Constants
    // ------------------------------------------------------------

    /** Simple empty URL[0] array. */
    public static final URL[] EMPTY_URL_ARRAY = new URL[0];

    // ------------------------------------------------------------
    //     Instance members
    // ------------------------------------------------------------

    /** The name of this loader. */
    private String name;

    // ------------------------------------------------------------
    //     Constructors
    // ------------------------------------------------------------

    /** Construct given a parent and a name.
     *
     *  @param parent The parent loader.
     *  @param name The name of this loader.
     */
    public ForeheadClassLoader(ClassLoader parent,
                               String name)
    {
        super( EMPTY_URL_ARRAY,
               parent );

        this.name = name;
    }

    // ------------------------------------------------------------
    //     Instance methods
    // ------------------------------------------------------------

    /** Retrieve the name of this ClassLoader.
     *
     *  @return The name of this loader.
     */
    public String getName()
    {
        return this.name;
    }

    /** Append a URL to this loader's search path.
     *
     *  @param url The URL to append.
     */
    public void addURL(URL url)
    {
        super.addURL( url );
    }

    // ------------------------------------------------------------
    //     java.lang.Object implementation
    // ------------------------------------------------------------

    /** Produce output suitable for debugging.
     *
     *  @return Output suitable for debugging.
     */
    public String toString()
    {
        return "[ForeheadClassLoader: name=\"" + getName() + "\"]";
    }

    /**
     * Find the resource with the given name, and return a String.
     * The search order is as described for <code>getResource()</code>,
     * after checking to see if the resource data has been previously cached.
     * If the resource cannot be found, return <code>null</code>.
     *
     * @param name Name of the resource to return a String for.
     */
    public String getResourceAsString( String name )
    {
        StringWriter sw = null;
        BufferedReader reader = null;

        try
        {
            sw = new StringWriter();
            InputStream is = getResourceAsStream( name );
            reader = new BufferedReader( new InputStreamReader( is ) );

            char buf[] = new char[1024];
            int len = 0;

            while ((len = reader.read(buf, 0, 1024)) != -1)
            {
                sw.write( buf, 0, len );
            }

            return sw.toString();
        }
        catch (IOException ioe)
        {
            return null;
        }
        finally
        {
            try
            {
                sw.close();
                reader.close();
            }
            catch (Exception ignored)
            {
            }
        }
    }

    /**
     * Add a new repository to the set of places this ClassLoader can look for
     * classes to be loaded.
     *
     * @param repository Name of a source of classes to be loaded, such as a
     *      directory pathname, a JAR file pathname, or a ZIP file pathname. The
     *      parameter must be in the form of an URL.
     * @exception IllegalArgumentException if the specified repository is
     *      invalid or does not exist
     */
    public void addRepository(String repository)
    {
        addRepository(new File(repository));
    }

    /**
     * Add a new repository to the set of places this ClassLoader can look for
     * classes to be loaded.
     *
     * @param repository Name of a source of classes to be loaded, such as a
     *      directory pathname, a JAR file pathname, or a ZIP file pathname. The
     *      parameter must be in the form of an URL.
     * @exception IllegalArgumentException if the specified repository is
     *      invalid or does not exist
     */
    public void addRepository(File repository)
    {
        if (repository.exists() && repository.isDirectory())
        {
            File[] jars = repository.listFiles();

            for (int j = 0; j < jars.length; j++)
            {
                if (jars[j].getAbsolutePath().endsWith(".jar"))
                {
                    try
                    {
                        URL url = jars[j].toURL();
                        super.addURL(url);                
                    }
                    catch (MalformedURLException e)
                    {
                        throw new IllegalArgumentException(e.toString());
                    }
                }
            }
        }
    }
}
